(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Absyn.sml,v
   Comment: High level abstract syntax based on wasowski/sestoft:2002.
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


   This is highlevel abstract syntax which is quite far away from concrete
   textual vsr syntax. It can be understood as IR suitable for interpretation
   or code generation, or perhaps abstract syntax of concrete visual syntax.
   It is much more conceptaul than textual description of statechart.

   CHANGES (comparing to VS):
   
   Grammar of exp, sexp and oexp and accesses has been generalized
   (orthogonalized). Arbitrary expressions may be now used as parameters,
   indexers etc.

   PROBLEMS:

   1. Variable ranges are simplified (not used currently)
   2. No Do activities
   3. No internal transitions (this should be resolved before 
   
   Absyn.sml,v 1.51 2002/11/09 10:23:55 wasowski Exp
*)

structure Absyn = struct

  open Types Haskellish Err Exp
  infixr 0 $

  
  exception StateNotFound;
  exception StateTypeMismatch;

  type symbol = Symbol.symbol
  type 'a env = 'a SymbolEnv.env
  
  val lookup  = SymbolEnv.lookup
  val bind1   = SymbolEnv.bind1
  val empty   = SymbolEnv.empty
  val bindZip = SymbolEnv.bindZip
  val isEmpty = SymbolEnv.isEmpty
  val toList  = SymbolEnv.toList

  type event = symbol
  type evgrpid = symbol
  type stateid = symbol	(* Assumed to be a fully qualified state name 
			   so state symbol identifies state uniquely:
			   systemname.topstatename. .... .myname *)

  type venv = value env

	  
  
  (* signal expression *)

  type sexp = event * exp list		    (* name, [param] *)
 
  (* actions *)

  type action = { assgns  : assgn list,
		  outputs : oexp list,
		  signals : sexp list }

  type actions = action list


  val empty_action = { assgns=[], outputs=[], signals=[] }

  (* states and hierarchy *)
 
  datatype marking = History | Initial 

  datatype orstate = ORState of {
			name     : stateid,			      
			parent   : stateid,
			children : andstate list,
			marking  : marking	  (* lifted up one level *)			      
		  }						      
		  
  and andstate = ANDState of {
			name     : stateid,
			parent   : stateid option,      (* NONE for root *)
			children : orstate list,
			entry    : action,
			exit     : action
		  }
  
  
  (* Union of state types. double constructors are unfortunately unavoidable
     because record types cannot be recursive themselves in SML97 *)

  datatype state = InclOR  of orstate
                 | InclAND of andstate

   (* In addition to state hierarchy I maintain state environment which is a
     flat index into hierarchy. This relies on the fact that hierarchy is
     persistent - static during execution. Similarly references to parent
     states could be embedded in the hierarchy in C style to allow movements
     up and down.  I kept the environment without being entirely convinced
     that I want it. Its main advantage is that the structure is not polluted
     with network of imperative references and all the processing remains
     pure. *)

  type stenv = state env

  (* hierarchy is modelled as the proper hierarchy starting with root state
     and an index to it *)

  type hierarchy = andstate * stenv
 
  fun getstate (h :hierarchy) (s :stateid) :state = lookup (#2 h) s  
  
  (* Combinators on functions acceping states *)

  fun lift_orand (fo :orstate -> 'a) (fa :andstate -> 'a) :state -> 'a =
	(fn s => case s of InclOR  so => fo so | InclAND sa => fa sa)
	
  fun lift_h (f :state -> 'a) : hierarchy -> stateid -> 'a =
	fn h => fn sid => f (getstate h sid) 
	
  fun lift_st (f : 'a -> 'b ) (g : 'a * 'b ->'b) = fn a => g (a, f a) 

  fun id (x :'a) :'a = x

  (* introduce slight inefficiencies but sometimes saves typing even more: *)

  fun lower_orand (f :state -> 'a) : (orstate -> 'a) * (andstate -> 'a) =
	(fn s => f (InclOR s), fn s => f (InclAND s))



  
  (* names of states *)
  
  fun nameor  (ORState  s) :stateid = #name s
  fun nameand (ANDState s) :stateid = #name s
  val name = lift_orand nameor nameand


 
  (* Comparison on states: mainly used to build efficient container types *)

  fun compareor (s1 :orstate, s2 :orstate) :order = 
	Symbol.compare (nameor s1,nameor s2)
	
  fun compareand(s1 : andstate, s2 :andstate) :order = 
	Symbol.compare (nameand s1,nameand s2)
	
  fun compare   (s1 : state, s2 :state) :order =
	Symbol.compare (name s1, name s2)


 
  structure StateSet = 
	SetFn (BinarySetFn( type ord_key= state val compare= compare ))

  structure ORStateSet =
	SetFn (BinarySetFn( type ord_key= orstate val compare= compareor ))

  structure ANDStateSet =
	SetFn (BinarySetFn( type ord_key= andstate val compare= compareand ))

  structure Set =
	SetFn (BinarySetFn( type ord_key= Symbol.symbol 
			    val compare= Symbol.compare ))
  
  type orstateset  = ORStateSet.set
  type andstateset = ANDStateSet.set
  type stateset    = StateSet.set
  type symbolset   = Set.set
  type eventset    = symbolset


  (* default state environment *)		  

  type denv = andstate env		  

  (* and to simple helpers to make constructing more succint *)
		 
  fun inor  st = InclOR  (ORState st)
  fun inand st = InclAND (ANDState st)
	

  (* index was not explicitly modelled in theoretical description of TR since
     it is only introduced to improve efficiency. The index is built by a
     function idx_hierarchy below. *)

  (* event bindings *)
  type ebind = symbol * symbol list	    (* name, [param_list] *)

  (* element types *)
  

  type evgroup = eventset 
  
  datatype trigger = EventGroup of symbol
		   | Event      of ebind 

  (* transition *)

  type guard = bexp list

  type tran = { trigger : trigger,
		scope   : orstate,
		pos     : andstate list,
		neg     : andstate list,
		guard   : guard,
		action  : action,
		targets : andstate list }
		 
  
  (* model *)
  

  type system = {

	name      : symbol,
	hierarchy : hierarchy,		(* the whole state business is here *)
	vartypes  : vartype  env,
	evtypes   : evtype   env,
	evgroups  : evgroup  env,
	otypes    : otype    env,
	denv      : denv,		(* mapping of default states        *)
	venv      : venv,		(* initial variable assignment      *)

	cenv      : const env,		(* named constants                  *)
					(* Presently I do not remember why  *)
					(* values and constants are not the *)
					(* same type. Perhaps there was no  *)
					(* reason ...  ?                    *)
					 
	trans     : tran list
  }


  type model = system list

  (* All types are done until now. Auxiliary functions follow *)

  fun getand (InclAND s) = SOME s | getand _ = NONE
  fun getor  (InclOR  s) = SOME s | getor  _ = NONE

  (* unboxing event trigger *)

  fun getebind (Event eb :trigger) = eb 
    | getebind (EventGroup _) = impossible "Event expected, group found!"


  fun gettree (h :hierarchy) :andstate = #1 h
  fun getroot (h :hierarchy) :state = InclAND (gettree h)



  fun parentor  (ORState s)  :stateid = #parent s
  fun parentand (ANDState s) :stateid = 
	case #parent s of SOME p => p | NONE => raise StateNotFound
  val parent  = lift_orand parentor parentand
  val parenth = lift_h parent 
  

  fun childrenand (ANDState s) :orstate list  = #children s
  fun childrenor  (ORState s)  :andstate list = #children s
  val children : state -> state list = 
	lift_orand (map InclAND o childrenor) (map InclOR o childrenand)

  fun entryand (ANDState s) :action = #entry s
  fun exitand  (ANDState s) :action = #exit s

  local
	fun dor (s :orstate, res :stateset) :stateset =
		let val kids = childrenor s
		    val res1 = foldl (fn(k,r)=>StateSet.add(r,InclAND k)) res kids
		in foldl dand res1 kids end

	and dand (s :andstate, res :stateset) :stateset = 
		let val kids= childrenand s
		    val res1= foldl (fn(k,r)=>StateSet.add(r,InclOR k)) res kids
		in foldl dor res1 kids end
  in
	fun descendor  (s :orstate ) :stateset = dor  (s,StateSet.empty)
	fun descendand (s :andstate) :stateset = dand (s,StateSet.empty)
  end
  val descendandst : andstate -> stateset =
	lift_st descendand (fn (s,ds) => StateSet.add(ds,InclAND s))
  val descendorst : orstate -> stateset =
	lift_st descendor  (fn (s,ds) => StateSet.add(ds,InclOR s))
  val descend  : state -> stateset = 
	lift_orand descendor descendand
  
  val descendst : state -> stateset = lift_st descend StateSet.add'
  
  val descendh : hierarchy -> stateid -> stateset = lift_h descend

  (* siblings - including self *)

  fun siblingsst (h :hierarchy) (s :state) :state list =
	children $ getstate h $ parent s

  fun siblingandst (h :hierarchy) (s :andstate) :andstate list =
	childrenor $ valOf $ getor $ getstate h $ parentand s
	handle Option => impossible "should always be orstate"
	
  fun isrootand (ANDState s) = #parent s = NONE
  val isroot    = lift_orand (fn _ => false) isrootand
  val isrooth   = lift_h isroot



  fun ishistoryor (ORState s) = #marking s = History

  (* entry/exit actions *)
 
  fun exand (ANDState s) = #exit  s
  fun enand (ANDState s) = #entry s
  val en = lift_orand  (fn _ => 
	Err.impossible "orstates have no entry actions") enand

  (* state types *)

  fun isand   (InclAND s) = true | isand _ = false
  fun isor    (InclOR  s) = true | isor  _ = false
  
  val isandh = lift_h isand
  val isorh  = lift_h isor
  
  
  fun isbasicand (ANDState s) = null (#children s)
  val isbasic = lift_orand (fn _ => false) isbasicand

 
  fun onlyand (s :stateset) :andstateset = 
	ANDStateSet.addList 
		(ANDStateSet.empty, List.mapPartial getand $ 
			StateSet.listItems s)

  fun onlyor (s :stateset) :orstateset =
	ORStateSet.addList
		(ORStateSet.empty, List.mapPartial getor $ 
			StateSet.listItems s)

    
  (* hierarchy constructor -- buids an index to explicit hierarchy starting
     with root *)  

  fun idx_hierarchy (root : andstate) : hierarchy = 
	let
		fun idxor stenv (ORState s) : stenv = 
			let 
				val stenv1 = bind1 stenv (#name s, inor s)
			in 
				foldl (fn (a,e) => idxand e a) 
					stenv1 (#children s) 
			end
		and idxand stenv (ANDState s) : stenv =
			let 
				val stenv1 = bind1 stenv (#name s, inand s)
			in	
				foldl (fn (a,e) => idxor e a) 
					stenv1 (#children s)
			end
	in (root, idxand empty root) end

  (* rebuild default environment after hierarchy was rebuilt. Unfortunately
     denv uses persistent references to andstates in hierarchy (instead of
     names) which makes model tranformations troublesome. This function takes
     an old denv and new (incompatible) hierarchy and builds new denv. It
     relies on the fact that state names are unique across the hierarchy! *)

  fun mkdenv (denv :denv) (root :andstate) :denv =
	let fun mkand (s :andstate, ndenv :denv) = 
			foldl mkor ndenv (childrenand s)
	    and mkor  (s :orstate,  ndenv :denv) =
			let val dname =	nameand (lookup denv (nameor s))
			    val dstop = List.find (fn s => dname = nameand s) 
						(childrenor s)
			    val dst   = valOf dstop handle Option => 
					impossible "there always is default st!"
			    val ndenv1= bind1 ndenv (nameor s,dst) 
			in foldl mkand ndenv1 (childrenor s) end
	in mkand (root, empty) end

  (* more properties of hierarchy *)

  (* list of ancestors is sorted with age decreasing -- root is the prefix *)	
  
  local 
	fun recancest h s ancs :stateid list =
		let val p = parenth h s 
		in recancest h p (p::ancs) end handle StateNotFound => ancs
  in
	fun ancesth (h :hierarchy) (s :stateid) :stateid list = 
		recancest h s []
		
	fun ancesthst (h :hierarchy) (s :stateid) :stateid list =    
		recancest h s [s]
  end




  (* checks if s1 is an ancestor of s2.  FIXME: could be more efficient
     traversing hierarchy and checking ancestors instead of computing anestors
     set in place *)

  fun ancestorh (h :hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	let val ancests = ancesth h s2
	in List.exists (fn a => a=s1) ancests end

  fun ancestorhst (h :hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	s1 = s2 orelse ancestorh h s1 s2



  (* returns true if s1 and s2 are ancestrally related *)

  fun ancestralh (h :hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	ancestorh h s1 s2 orelse ancestorh h s2 s1

  fun ancestralhst (h :hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	ancestorhst h s1 s2 orelse ancestorhst h s2 s1
  

 
  (* Lowest Common Ancestor *) 

  (* Stephen and Christian have much better solution for the problem (which
     is said to be "optimal"). Note also that alternatively this can be
     defined for a pair of states and then folded. For practical reasons this
     may be a bit faster. Also we usually need it for a pair. *)

  fun LCA (h :hierarchy) [] :stateid = 
		impossible "LCA is not defined for empty set"
    | LCA (h :hierarchy) (ss :stateid list) :stateid =
	let	
		val a = map (ancesthst h) ss 
		fun f (l :stateid list list) (lca :stateid) =
			let val hds = map hd l 
			    val fst = hd hds 
			in 
				if List.all (fn h => h=fst) hds 
				then f (map tl l) fst else lca
			end handle Empty => lca
	in f a (name $ getroot h) end

  (* irreflexive version of LCA - it is guaranteed that result is a strong
     parent of all list elements. Undefined for any list containing topstate. *)

  fun irrLCA (h :hierarchy) [] :stateid = 
		impossible "irrLCA is not defined for empty set"
    | irrLCA (h :hierarchy) (ss :stateid list) :stateid =
	let val a = map (ancesth h) ss 
	    fun f (l :stateid list list) (lca :stateid) =
		let val hds = map hd l 
		    val fst = hd hds 
		in if List.all (fn h => h=fst) hds 
		   then f (map tl l) fst else lca
		end handle Empty => lca
	in f a (name $ getroot h) end

  (* the most trivial and unefficient algorithm for testing orthogonality *)

  fun orthogonal (h :hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	(not (ancestralh h s1 s2)) andalso isandh h (LCA h [s1,s2])

  (* events set of transitions trigger *)	

  fun events (eg :evgroup env) (t :tran) :eventset =
	case #trigger t of
		  Event e	=> Set.singleton (#1 e)
		| EventGroup e	=> lookup eg e

  
  (* rightmost descendant (RMD) of given orstate/andstate *)

  fun RMDor (s :orstate) :andstate = 
	let val kid = List.last (childrenor s)
	in if null (childrenand kid) then kid
	   else RMDand kid
	end

  and RMDand (s :andstate) :andstate =
	let val _ = assert (fn ()=> not (null (childrenand s)))
			"You should never need to ask for RMD of basic state"
	in RMDor (List.last (childrenand s)) end

	

  
  (* another weird bunch of experiments with higher order programming *)

  (* general fold over hierarchy.  This fold has a starting point e
     used for children of basic states. So actually fa(s,e) is called if s is a
     basic state.
     This fold applies fa and fo function on the hierarchy in pre-order *)

  fun hfoldor (fo :orstate * 'a -> 'a) (fa :andstate * 'a -> 'a) (e : 'a)
	       (s :orstate ) : 'a = 
       fo (s,foldr (fn (s',e')=> hfoldand fo fa e' s') e (childrenor s))

  and hfoldand (fo :orstate * 'a -> 'a) (fa :andstate * 'a -> 'a) (e : 'a)
	       (s :andstate) : 'a = 
       fa (s,foldr (fn (s',e')=> hfoldor  fo fa e' s') e (childrenand s))

  fun hfold (fo :orstate * 'a -> 'a) (fa :andstate * 'a -> 'a) (e :'a)
		(s :state) :'a =
	case s of InclOR  s => hfoldor fo fa e s
		| InclAND s => hfoldand fo fa e s

  (* a left to right version of the above *)

  fun hfoldlor (fo :orstate * 'a -> 'a) (fa :andstate * 'a -> 'a) (e : 'a)
	       (s :orstate ) : 'a = 
       fo (s,foldl (fn (s',e')=> hfoldland fo fa e' s') e (childrenor s))

  and hfoldland (fo :orstate * 'a -> 'a) (fa :andstate * 'a -> 'a) (e : 'a)
	       (s :andstate) : 'a = 
       fa (s,foldl (fn (s',e')=> hfoldlor  fo fa e' s') e (childrenand s))

  fun hfoldl (fo :orstate * 'a -> 'a) (fa :andstate * 'a -> 'a) (e :'a)
		(s :state) :'a =
	case s of InclOR  s => hfoldlor fo fa e s
		| InclAND s => hfoldland fo fa e s

  (* sort transitions by priority. To achieve VS conflict resoltuion
     states are assigned priorities in preorder traverse and transitions are
     sorted by priorities of their scopes.  FIXME: Inefficient. For instance
     priorites could be explicitly kept in hierarchy (computed while it is
     being built) and then just sorted using Listsort. This inefficiency does
     not influence interpreting speed (and generated code). Function is only
     called once - during translation from vsr -> Absyn *)

  fun prioritize (root :andstate) (trans :tran list) :tran list =
	let
		fun fo (s :orstate, tr) :tran list = 
			List.filter (fn t => #scope t = s) trans @ tr
		fun fa (s :andstate,tr) :tran list = tr
		val prioritized = hfoldand fo fa [] root 
		val _ = assert (fn () => length trans = length prioritized)
			"Some transitions has been lost during sorting."
	in prioritized end

end
