(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/AbsynLogic.sml,v
   Comment: Analysing Absyn models with BDDs
   Release: scope-0-10
     
     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   This module can only work with one model at a time. This is restriction of
   BUDDY package, which without special cleverness can only on one problem at
   a time (there is a signel variable pool). This can be overcome quite
   easily, but seems to be not needed currently.

   THIS MODULE USES IMPERATIVE FEATURES!!!

   AbsynLogic.sml,v 1.12 2002/10/24 20:17:09 wasowski Exp
*)

structure AbsynLogic :> AbsynLogic = struct

  open Haskellish Absyn Say Err AbsynPrint
  infixr 0 $
  infix 9 sub

  val all = List.all

  type 'a array = 'a Array.array
  type bdd = bdd.bdd
  type varnum = bdd.varnum
  type varSet = bdd.varSet

  val state2i_env : varnum env ref = ref SymbolEnv.empty;
  val i2state_env : stateid array ref = ref (Array.fromList []);

  fun state2i (s :stateid) :varnum = lookup (!state2i_env) s 
  fun i2state (i :varnum) :stateid = Array.sub(!i2state_env,i)

  (* print mappings of state names to bdd variable numbers.
     used for testing and debugging *)
     
  fun sprMapping () :string =
	let val os : string list ref = ref []
	    fun pr(i,s) = 
		os:=(Int.toString i ^ " -> " ^ Symbol.name s ^ "\n")::(!os)
	in 
		Array.appi pr (!i2state_env,0,NONE); 
		String.concat (!os)
	end

  fun sprSolution (pos,neg) :string = 
	let val p = String.concat $ Typeset.join_blanks(map Symbol.name pos) []
	    val n = String.concat $ Typeset.join_blanks(map Symbol.name neg) []
	in String.concat ["POS(",p,") NEG(", n, ")"] end

  (* variables not in the domain are ignored *)

  fun states2is (ss :stateid list) :varnum list =
	foldl (fn (s,ss) => state2i s::ss handle Subscript => ss) [] ss


  fun release () :unit = (

	state2i_env := SymbolEnv.empty;
	i2state_env := Array.fromList []
  )

  fun realloc (n :int) :unit = 
	if bdd.getVarnum() < n then bdd.setVarnum n else ()

  (* one ooperator missed in muddy *)

  fun IMPLIES (x :bdd, y:bdd) :bdd = bdd.OR(bdd.NOT x,y)

  (* returns the number of solutions as an integer - you should take care
     that there is no overflow. Returns the number of satisfying assignments
     only for the support. A bit around the world but I do not have anything
     else. 
     
     This is only for debugging purposes. Does not work for arbitrary big
     spaces. *)

  fun satcount_supp (r :bdd) :int =
	let val n = bdd.getVarnum() - (Vector.length $ bdd.scanset $ bdd.support r)
	    fun res x 0 = x
	      | res x n = res (x/2.0) (n-1)
	in Real.floor $ res (bdd.satcount r) n end 
	handle Overflow => ~1
		
		


  (* FIXME? This function imposes an order on variables in bdd. The order is
     slightly weird as the top variables are leaves of the statecharts. IT
     might be that the tree would become much more regular if the order is
     changed. At the moment this does not influence operation, but it might
     become important for big models. AW 05/07/2002 *)

  fun build_var_bindings (root :state) :unit =
	let val _ = state2i_env := empty
	    fun fo (s :orstate,  (i,l)) = 
		let val _ = (state2i_env := bind1 (!state2i_env) (nameor s,i))
		in (i+1,nameor s :: l) end
	    fun fa (s :andstate, (i,l)) = 
		let val _ = (state2i_env := bind1 (!state2i_env) (nameand s,i))
		in (i+1, nameand s::l) end
	    val (varnum,l) = hfold fo fa (0,[]) root 
	in 
		i2state_env := Array.fromList (rev l);
		realloc varnum 
	end

   fun restrict_to (r :bdd) (ss :stateid list) :bdd =
	let val ss1	= states2is ss 
	    val allvars = List.tabulate(bdd.getVarnum(),fn x=>x)  
	    fun f v	= all (fn s => v<>s) ss1
	    val delete	= List.filter f allvars
	    val r1 = bdd.exist (bdd.makeset delete) r
	in IfVerbose( fn () => (
		print "Existential complementing: ";
		app (fn i=> printi i before print " ") ss1;
		print "=> E.qualifying: ";
		app (fn i => printi i before print " ") delete;
		print "\nsupport left: ";
		Vector.app (fn i => print (Int.toString i ^ " ")) 
				(bdd.scanset $ bdd.support r1);
		cr()
	   ));
	   r1
	end handle Subscript => impossible "#19"

  (* below are three forcing functions. Forcing takes a bdd describing a set
     of configurations and returns a bdd describing a subset which
     additionally satisfies given list of positive and negative conditions.
     Note that this is implemented using bdd.AND, not restrict. Initial
     implementation with restrict was wrong. There is a subtle difference
     between and and restrict in bdds. Both operations would result in the
     same set of solutions (modulo number of variables). The difference is
     that restrict eliminates variables from bdd (and its support), while AND
     forces them to be true (and they remain in support). We need sometimes to
     refer to positive conditions (for instance when a transition goes down
     the hierarchy), so this variables should still be present in the solution
     set. This is way AND is used. AW 6/07/2002 *)

  (* variables not in the domain are ignored *)

  fun force_pos (r :bdd) (pos :stateid list) :bdd = 
	let val ass = bdd.assignment $ map (fn s => (s,true)) $ states2is pos
	in bdd.AND(bdd.fromAssignment ass, r) end

  (* variables not in the domain will be ignored *)

  fun force_neg (r :bdd) (neg :stateid list) :bdd =
	let val ass = bdd.assignment $ map (fn s => (s,false)) $ states2is neg
	in bdd.AND (bdd.fromAssignment ass, r) end

  (* variables not in the domain will be ignored *)

  fun force (r :bdd) (pos :stateid list,neg :stateid list) :bdd =
	let val pos1 = map (fn s => (s,true)) $ states2is pos
	    val neg1 = map (fn s => (s,false)) $ states2is neg
	    val ass  = bdd.assignment (pos1 @ neg1)
	in IfVerbose( fn()=> (
		sayln Verbose ("Forcing pos: " ^ sprInts (map #1 pos1));
		sayln Verbose ("Foricng neg: " ^ sprInts (map #1 neg1)) ));
		bdd.AND(bdd.fromAssignment ass, r)
	end

  (* get one satisfying assignment and exclude it from bdd. See comment
     by allsat for explanation of vars param. vars is assumed to contain
     no repetitions *)

  fun nextsat (r :bdd) (vars :stateid list) 
		:(stateid list * stateid list) list * bdd =
	let val ass = bdd.satone r
	    val r1 = bdd.AND(bdd.NOT(bdd.fromAssignment ass),r)
	    val (pos,neg) = List.partition #2 (bdd.getAssignment ass)
	    val pos1 = map (i2state o #1) pos
	    val neg1 = map (i2state o #1) neg
	    fun fresh s = all (fn s1 => s1 <> s) pos1 andalso
			  all (fn s1 => s1 <> s) neg1 
	    val fvars = List.filter fresh vars
	    fun extend s ((p,n),ass) = (p,s::n)::(s::p,n)::ass
	    fun mult []      ass = ass
	      | mult (s::ta) ass = mult ta (foldl (extend s) [] ass)
	in (mult fvars [(pos1,neg1)],r1) end

  (* vars is the set of variables which *must* be included in solutions
     returned. All assignements returned refer to vars union support(r). 
     To get result on support only use allsat r []. The overhead of processing 
     empty list is very small. vars may contain repetitions which will be
     filtered out. *)

  fun allsat (r :bdd) (vars :stateid list) :(stateid list * stateid list) list =
	let val vars1 = Set.listItems $ Set.fromList vars 
	    fun f (r :bdd) acc =
		if bdd.equal r bdd.FALSE then acc
		else let val (lst,r1) = nextsat r vars1 in f r1 (lst @ acc) end
	in f r [] end

  fun buildand (s :andstate) =
	let val e = bdd.ithvar (state2i (nameand s))
	    val ne = bdd.NOT e
	    val kids = map (bdd.ithvar o state2i o nameor) (childrenand s)
	    val kidsand = foldl bdd.AND bdd.TRUE kids
	    val nkids   = map bdd.NOT kids
	    val nkidsand= foldl bdd.AND bdd.TRUE kids
	    val r = foldl (fn (s,d)=> (bdd.AND(buildor s,d))) 
			bdd.TRUE (childrenand s) 
	    val _ = IfVerbose(fn()=> bdd.fnprintdot ((Symbol.name $ nameand s)^".dot" ) r)
	in bdd.AND(bdd.AND(IMPLIES(e,kidsand),IMPLIES(ne,nkidsand)), r) end

  and buildor (s :orstate) =
	let val e       = bdd.ithvar (state2i (nameor s))
	    val ne      = bdd.NOT e
	    val kids    = map (bdd.ithvar o state2i o nameand) (childrenor s)
	    val kidsxor = foldl bdd.XOR bdd.FALSE kids
	    val nkids   = map bdd.NOT kids
	    val nkidsand= foldl bdd.AND bdd.TRUE nkids
	    val r       = foldl (fn (s,d) => (bdd.AND(buildand s,d)))
			  bdd.TRUE (childrenor s)
	in bdd.AND(bdd.AND(IMPLIES(e,kidsxor),IMPLIES(ne,nkidsand)), r) end

  fun build (s :state) :bdd = 
	let val _ = build_var_bindings s
	    val r = (lift_orand buildor buildand) s
	in bdd.AND( bdd.ithvar $ state2i $ name s, r ) end

  fun build_sys (sys :system) :bdd =
	build $ InclAND $ gettree $ #hierarchy sys

  fun done () = (

	bdd.done();
	release()
  )

  

end
