(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Main.sml,v
   Comment: Main source file for SCOPE
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   TODO: 1. Create Main.sig

   2. anonymizing and and simple parse/save should be made into one code
   generator: vsr.  This is a lot of work since the Unparse module does not
   use the Absyn format. Moreover Absyn would loose all layout information.
   Anonymizing should just be a model transformation option applicable to
   this (and all other) code generators. Similarly other model
   transformations should be applicable to this one here. But its quite a lot
   of work to port Unparse to Absyn.

   3. Processing of actions should be changed. Firstly the model should be
   parsed. Secondly model transformations should be applied in the order
   specified on the command line. Finally Cod Generator should be called.
   Of course anonymizer does not really fit well in the scheme now.

   Main.sml,v 1.61 2002/11/11 22:16:11 wasowski Exp
*)

structure Main = struct

  (* some interface *)

  open Say Haskellish;
  infixr 0 $;

  val parse = parse.parse			      
  structure G = GetOpt
  
  (* these are defaults for some options *)
  (* imperative but only used locally inside this module *)

  val odir     : string ref = ref "./"
  val ofile    : string option ref = ref NONE
  val cg_filert: string option ref = ref NONE
  val stubmode : CodGenC.stubsty option ref  = ref NONE
  val drvmode  : CodGenC.driverty option ref = ref NONE
  val smoke    : bool ref = ref false
  val debugmode: bool ref = ref false
  val queuelen : int ref  = ref 10
  val externhdr: string option ref = ref NONE

  (* Print out statistics. Should it be here? *)

  fun statsys (sys :Absyn.system) : unit =
	let val nam  = "system:" ^ Symbol.name (#name sys)
	    val ss   = " states:" ^ Int.toString (Stat.states sys)
	    val ans  = " and:" ^ Int.toString (Stat.and_states sys)
	    val bs   = " basic:" ^ Int.toString (Stat.basic_states sys)
	    val his  = " his:" ^ Int.toString (Stat.history_sts sys)
	    val cb   = " confbound:" ^ Int.toString (Stat.conf_card_bound sys)
	    val d    = " depth(and):" ^ Int.toString (Stat.depth sys)
	    val rors = " rors:" ^ Int.toString (Stat.rors sys)
	    val tran = " tran:" ^ Int.toString (Stat.ntran sys)
	    val acgd = " tran_acgd:" ^ Int.toString (Stat.ntran_acgd sys)
	    val noac = " tran_noac:" ^ Int.toString (Stat.ntran_noac sys)
	    val nogd = " tran_nogd:" ^ Int.toString (Stat.ntran_nogd sys)
	    val tneg = " tran_neg:" ^ Int.toString (Stat.ntran_neg sys)
	    val (en,ex) = Stat.count_enex (#hierarchy sys)
	    val enex = " en/ex: " ^ Int.toString en ^ "/" ^ Int.toString ex 
	    val (sen,sex) = Stat.count_sigenex (#hierarchy sys)
	    val senex= " sig-en/ex: " ^ Int.toString sen ^ "/" ^ Int.toString sex
	    val (ev,si) = Stat.eventno sys 
	    val evsi = " ev/si: " ^ Int.toString ev ^ "/" ^ Int.toString si
	in 
		sayln VNorm (nam ^ ss ^ ans ^ bs ^ his ^ cb ^ d 
			^ rors ^ tran ^ noac ^ nogd ^ acgd ^ tneg ^ enex
				^ senex ^ evsi) 
	end

(* FIXME: satc only computes proper number of legal configurations
   for single system models. The results are much over saturated otherwise *)

  fun statsys2 (sys: Absyn.system) : unit =
	let val _ = statsys sys
	    val r = AbsynLogic.build_sys sys 
	    val satc = bdd.satcount r
	    val confcard = " conf-card:" ^ Real.toString(satc)
	    val scopes = Stat.get_scopes sys
	    val scono = " scopes: " ^ Int.toString (length scopes)
	    val scoroots = Stat.roots (#hierarchy sys) (map Absyn.InclOR scopes)
	    val rono  = " root-scopes: " ^ Int.toString (length scoroots)
	    val expuresc = List.filter Stat.is_expureor scopes
	    val expur = " expure-scopes: " ^ Int.toString (length expuresc)
	in  sayln VNorm (confcard ^ scono ^ rono ^ expur)  end
	
  fun stat (fname :string) (m :Absyn.model) (expensive :bool) :unit =
	let val _ = sayln VNorm ("-- file: " ^ fname) 
	in
		if expensive then  app statsys2 m
		else  app statsys m
	end 

(* auxiliaries *)

  (* runindir. warning: also creates a directory *)

  fun runindir (odir : string) (f : unit -> 'a) : 'a =
	let val curdir = FileSys.getDir()				
	    val _      = (						
				if FileSys.isDir odir then ()		
				else raise Fail ("Entry with name "	
				^ odir ^ " already exists and is \	
				\not a directory.")			
			 )						
			 handle SysErr _ => FileSys.mkDir odir		
	    val _      = FileSys.chDir odir				
	    val result = f()						
	in								
		FileSys.chDir curdir;					
		result							
	end								
		
	fun getAbsyn ifile :Absyn.model =
		VsrAbsyn.translate (parse ifile NONE)

	fun vsr2vsr ifile odir = 
		let val ast = parse ifile NONE
		in  runindir odir (fn () => Unparse.save ast) end

	fun anonymize ifile odir = 
		let
			val ast   = parse ifile NONE
			val _	  = Symbol.anonymize true
			val _	  = runindir odir (fn ()=>Unparse.save ast)
		in Symbol.anonymize false end

(* code generators *)

fun codgenc1 ifile m :unit =
	let val compiler = fn ()=>CodGenC1.compile 
		(!debugmode, !queuelen, !externhdr, !ofile) ifile m
	in	
		cg_filert := !ofile;
		runindir (!odir) compiler
	end

fun codgencstubs ifile m :unit =
	runindir (!odir) (fn ()=>CodGenCstubs.compile 
		(!debugmode,!ofile,false) ifile m)

fun codgenvsstubs ifile m :unit =
	runindir (!odir) (fn ()=>CodGenCstubs.compile
		(!debugmode,!ofile,true) ifile m)

fun codgencdrv ifile m :unit =
	runindir (!odir) (fn ()=>CodGenCdrv.compile 
		(!debugmode,!smoke,false,!ofile,!cg_filert) ifile m)

fun codgenvsdrv ifile m :unit =
	runindir (!odir) (fn ()=>CodGenCdrv.compile
		(!debugmode,!smoke,true,!ofile,!cg_filert) ifile m)

fun codgenscope ifile m :unit =
	runindir (!odir) (fn ()=>CodGenScope.compile (!ofile) ifile m)

  (* user interface *)

  datatype flag = 
	      Anonymize 
	    | Vsr2Vsr 
	    | OutDir	of string
	    | OutFile	of string 
	    | ExternHdr of string
	    | CodGen	of string
	    | Stats
	    | FullStats
	    | Version
	    | Verbosity	of verbosity
	    | Help
	    | DebugMode
	    | ReleaseMode
	    | SmokeMode
	    | QueueLen	of int

  fun cgmode (SOME m) = CodGen m
    | cgmode NONE     = CodGen "C1"
 
  fun mk_queuelen (s :string) = 
	case Int.fromString s of
		  SOME n => if n > 0 then QueueLen n
			    else Err.wrongInput "Queue length must be positive"
		| NONE   => Err.wrongInput "Queue length must be integer"

  val options = 
	[{short="a",
	  long=["anonymize"],
	  desc=G.NoArg Anonymize,
	  help="anonymize vsr/vsp file/project"},
	 {short="v",
	  long=["parsevsr"],
	  desc=G.NoArg Vsr2Vsr,
	  help="parse and save vsr/vsp file"},
	 {short="d",
	  long=["outdir"],
	  desc=G.ReqArg(OutDir,"OUTDIR"),
	  help="output directory (default: .)"},
	 {short="o",
	  long=["outfile"],
	  desc=G.ReqArg (OutFile,"OUTFILE"),
	  help="name of the output file"},
	 {short="c",
	  long=["codgen"],
	  desc=G.OptArg (cgmode, "MODE"),
	  help="compile in mode (default: C1)]\n\
		\\t\t\t\t\tC1     - normal code generator targetting C\n\
		\\t\t\t\t\t         (affected by debug/release/outfile)\n\
		\\t\t\t\t\tCstubs - generates stub functions for actions\n\
		\\t\t\t\t\t         (affected by debug/release/outfile)\n\
		\\t\t\t\t\tCdrv   - generates driver loop for actions \n\
		\\t\t\t\t\t         (affected by debug/release/outfile)\n\
		\\t\t\t\t\tVSdrv  - like Cdrv but visualSTATE compatible\n\
		\\t\t\t\t\tVSstubs- like Cstubs but visualSTATE compatible\n\
		\\t\t\t\t\tscope  - saves model in internal format."},
	 {short="t",
	  long=["debug"],
	  desc=G.NoArg DebugMode,
	  help="generate code with test checks (debug outputs, etc),\n\
	       \\t\t\t\t\tassertions,symbolic event names, etc\n\
	       \\t\t\t\t\tMode of generation (release/debug) may be changed\n\
	       \\t\t\t\t\tbetween code generators, but not all combinations\n\
	       \\t\t\t\t\tare compatible."},
	 {short="r",
	  long=["release"],
	  desc=G.NoArg ReleaseMode,
	  help="generate code with no debugging overheads, \n\
	       \\t\t\t\t\tno printouts, etc (default)"},
	 {short="",
	  long=["smoke"],
	  desc=G.NoArg SmokeMode,
	  help="Use smoke driver instead of tiny driver or debug driver."},
	 {short="l",
	  long =["queuelen"],
	  desc=G.ReqArg(mk_queuelen, "LENGTH"),
	  help="number of entries reserved in signal queue.\n\
	       \\t\t\t\t\tOnly relevant in code generation (C1).\n\
	       \\t\t\t\t\tNote: you can learn it from verificator."},
	 {short= "",
	  long = ["externhdr"],
	  desc = G.ReqArg(ExternHdr, "FILE.h"),
	  help = "header file containing prototypes for external\n\
		  \\t\t\t\t\tfunctions used in actions and guards. \n\
		  \\t\t\t\t\t#included by code generator (C1). If absent\n\
		  \\t\t\t\t\tsystemname_extern.h is assumed.\n"},
	 {short="s",
	  long=["stats"],
	  desc=G.NoArg Stats,
	  help="get model statistics"},
	 {short="S",
	  long=["fullstats"],
	  desc=G.NoArg FullStats,
	  help="full statistics (potentially slow)"},
	 {short="",
	  long=["version"],
	  desc=G.NoArg Version,
	  help="get release name of your scope"},
	 {short="V",
	  long=["verbosity"],
	  desc=G.ReqArg(Verbosity o fromInt o valOf o Int.fromString, 
	       "LEVEL"),
	  help="0 - full debug, 7 - quiet (default 5)"},
	 {short="",
	  long=["quiet"],
	  desc=G.NoArg (Verbosity (Say.fromInt 7)),
	  help="equivalent to --verbosity=7"},
	 {short="h",
	  long=["help","usage"],
	  desc=G.NoArg Help,
	  help="display this usageinfo"}];

  (* filter out all non-command flags, executing their semantic actions 
   **)

  fun prepflags [] = []
    | prepflags (OutDir    d::ta) = (odir := d;           prepflags ta)
    | prepflags (ExternHdr f::ta) = (externhdr := SOME f; prepflags ta)
    | prepflags (QueueLen n::ta)  = (queuelen  := n;      prepflags ta)
    | prepflags (f::ta)		  = f :: prepflags ta

  (* run those flags which are commands *)

  fun runflag ifile (Anonymize,m :Absyn.model) :Absyn.model = 
		m before (anonymize ifile (!odir))
		(* FIXME: przerobic, aby korzystao z Absyn *)
    | runflag ifile (Vsr2Vsr,m) = 
		m before (vsr2vsr ifile (!odir))
		(* FIXME: przerobic, aby korzystao z Absyn *)
    | runflag ifile (SmokeMode  ,m) = m before (smoke := true)
    | runflag ifile (OutFile f  ,m) = m before (ofile := SOME f)
    | runflag ifile (Verbosity l,m) = m before (setverbosity l)
    | runflag ifile (Stats      ,m) = m before (stat ifile m false)
    | runflag ifile (FullStats  ,m) = m before (stat ifile m true)
    | runflag ifile (DebugMode  ,m) = m before (debugmode := true)
    | runflag ifile (ReleaseMode,m) = m before (debugmode := false)

    | runflag ifile (CodGen cg,m)   = 
	m before (case cg of
		 "C1"     => codgenc1      ifile m
	       | "Cstubs" => codgencstubs  ifile m 
	       | "Cdrv"   => codgencdrv    ifile m
	       | "VSdrv"  => codgenvsdrv   ifile m
	       | "VSstubs"=> codgenvsstubs ifile m
	       | "scope"  => codgenscope   ifile m
	       | _        => Err.wrongInput ("Unknown code generator mode:" ^ cg))
    | runflag _ _ = Err.impossible "Attempt to run a non-command flag!"
   
  val header =	"\nSCOPE -- StateCharts cOmPilEr release scope-0-10\n\
		 \Andrzej Wasowski. IT-University. Copenhagen, 2001-2002\n\
		 \usage: scope [OPTIONS] input_file\n"

  val note   =	"\nMost options can be repeated. The last occurence before a\n\
		 \command (-c) determines the behaviour of the command (with\n\
		 \nrespect to the specific option).\n\n\
		 \This is an experimental tool wih no full end-user support.\n\
		 \Use it with care and at your own responsibility.\n\
		 \Multiply nested directories are not created.\n\  
		 \EXISTING FILES ARE SILENTLY OVERWRITTEN in destination directory.\n\n"

  fun bad_cmdline errs = print "\n" 
       before app (say VLow) errs 
       before sayln VNorm (G.usageInfo header options)
       before sayln VNorm note

  fun runflags (flags,inputs,[]) = 
	if List.exists (fn f => f = Help) flags then
		sayln VNorm (G.usageInfo header options)
		before sayln VNorm note
	else if List.exists (fn f => f = Version) flags then
		sayln VNorm "scope-0-10"
	else if null inputs then
		bad_cmdline ["scope: No input file\n"]
	else if length inputs > 1 then
		bad_cmdline ["Only one input file is accepted.\n"]
	else (
		case flags of Verbosity l::ta => setverbosity l
			    | _ => ();
		
		bdd.init 10000 10000;
		foldl (runflag (hd inputs)) (getAbsyn (hd inputs)) (prepflags flags);
		bdd.done() 
	)

    | runflags (_,_,errs) = bad_cmdline errs
  
  (* main: initialize the compiler etc *)

  fun main() =
          runflags (G.getOpt G.Permute options (CommandLine.arguments()))
	  handle Fail s => Location.errPrompt s before
				 Location.errPrompt "Scope terminated."
	       |      e => Err.exnhandler e (fn ()=>())

end
